@@ -48,10 +48,25 @@ use vars qw(@ISA $VERSION $DEBUG $SSL_ERROR $GLOBAL_CONTEXT_ARGS @EXPORT );
 	@EXPORT = qw( SSL_WANT_READ SSL_WANT_WRITE $SSL_ERROR GEN_DNS GEN_IPADD );
 }
 
+my @caller_force_inet4; # in case inet4 gets forced we store here who forced it
+my $can_ipv6;           # true if we successfully enabled ipv6 while loading
+
 BEGIN {
 	# Declare @ISA, $VERSION, $GLOBAL_CONTEXT_ARGS
-	@ISA = qw(IO::Socket::INET);
-	$VERSION = '1.13_5';
+	
+	# if we have IO::Socket::INET6 we will use this not IO::Socket::INET, because
+	# it can handle both IPv4 and IPv6. If we don't have INET6 available fall back
+	# to INET
+	if ( ! eval {
+		require Socket6;
+		Socket6->import( 'inet_pton' );
+		require IO::Socket::INET6;
+		@ISA = qw(IO::Socket::INET6);
+		$can_ipv6 = 1;
+	}) {
+		@ISA = qw(IO::Socket::INET);
+	}
+	$VERSION = '1.16_1';
 	$GLOBAL_CONTEXT_ARGS = {};
 
 	#Make $DEBUG another name for $Net::SSLeay::trace
@@ -67,7 +82,7 @@ BEGIN {
 }
 
 sub DEBUG {
-	$DEBUG or return;
+	$DEBUG>=shift or return; # check against debug level
 	my (undef,$file,$line) = caller;
 	my $msg = shift;
 	$file = '...'.substr( $file,-17 ) if length($file)>20;
@@ -113,13 +128,22 @@ sub import {
 	my @export;
 	foreach (@_) { 
 		if ( /^inet4$/i ) {
-			require IO::Socket::INET;
-			@ISA = 'IO::Socket::INET'
+			# explicitly fall back to inet4
+			@ISA = 'IO::Socket::INET';
+			@caller_force_inet4 = caller(); # save for warnings for 'inet6' case
 		} elsif ( /^inet6$/i ) {
-			require IO::Socket::INET6;
-			require Socket6;
-			Socket6->import( 'inet_pton' );
-			@ISA = 'IO::Socket::INET6'
+			# check if we have already ipv6 as base
+			if ( ! UNIVERSAL::isa( $class, 'IO::Socket::INET6' )) {
+				# either we don't support it or we disabled it by explicitly
+				# loading it with 'inet4'. In this case re-enable but warn
+				# because this is probably an error
+				if ( $can_ipv6 ) {
+					@ISA = 'IO::Socket::INET6';
+					warn "IPv6 support re-enabled in __PACKAGE__, got disabled in file $caller_force_inet4[1] line $caller_force_inet4[2]";
+				} else {
+					die "INET6 is not supported, missing Socket6 or IO::Socket::INET6";
+				}
+			}
 		} elsif ( /^:?debug(\d+)/ ) {
 			$DEBUG=$1;
 		} else {
@@ -199,7 +223,8 @@ sub configure_SSL {
 
 	#Handle CA paths properly if no CA file is specified
 	if ($arg_hash->{'SSL_ca_path'} ne '' and !(-f $arg_hash->{'SSL_ca_file'})) {
-		warn "CA file $arg_hash->{'SSL_ca_file'} not found, using CA path instead.\n" if ($DEBUG);
+		DEBUG(1, "CA file $arg_hash->{SSL_ca_file} not found, using CA path instead.\n" )
+			if $arg_hash->{SSL_ca_file};
 		$arg_hash->{'SSL_ca_file'} = '';
 	}
 
@@ -224,9 +249,12 @@ sub configure_SSL {
 			$host or return $self->error( "Cannot determine peer hostname for verification" );
 
 			# verify name
-			DEBUG( "check $host with $vcn_scheme against $cert" );
 			my $x509 = Net::SSLeay::X509_STORE_CTX_get_current_cert($ctx_store);
-			return verify_hostname_of_cert( $host,$x509,$vcn_scheme );
+			my $rv = verify_hostname_of_cert( $host,$x509,$vcn_scheme );
+			# just do some code here against optimization because x509 has no
+			# increased reference and CRYPTO_add is not available from Net::SSLeay
+			DEBUG(99999,"don't to anything with $x509" );
+			return $rv;
 		};
 	}
 
@@ -262,9 +290,9 @@ sub connect {
 		# if this fails this might not be an error (e.g. if $! = EINPROGRESS
 		# and socket is nonblocking this is normal), so keep any error
 		# handling to the client
-		#DEBUG( 'socket not yet connected' );
+		DEBUG(2, 'socket not yet connected' );
 		$self->SUPER::connect(@_) || return;
-		#DEBUG( 'socket connected' );
+		DEBUG(2,'socket connected' );
 	}
 	return $self->connect_SSL;
 }
@@ -277,7 +305,7 @@ sub connect_SSL {
 	my ($ssl,$ctx);
 	if ( ! ${*$self}{'_SSL_opening'} ) {
 		# start ssl connection
-		#DEBUG( 'ssl handshake not started' );
+		DEBUG(2,'ssl handshake not started' );
 		${*$self}{'_SSL_opening'} = 1;
 		my $arg_hash = ${*$self}{'_SSL_arguments'};
 
@@ -308,7 +336,7 @@ sub connect_SSL {
 		? $args->{Timeout} 
 		: ${*$self}{io_socket_timeout}; # from IO::Socket
 	if ( defined($timeout) && $timeout>0 && $self->blocking(0) ) {
-		#DEBUG( "set socket to non-blocking to enforce timeout=$timeout" );
+		DEBUG(2, "set socket to non-blocking to enforce timeout=$timeout" );
 		# timeout was given and socket was blocking
 		# enforce timeout with now non-blocking socket
 	} else {
@@ -320,17 +348,17 @@ sub connect_SSL {
 	for my $dummy (1) {
 		#DEBUG( 'calling ssleay::connect' );
 		my $rv = Net::SSLeay::connect($ssl);
-		#DEBUG( "connect -> rv=$rv" );
+		DEBUG( 3,"Net::SSLeay::connect -> $rv" );
 		if ( $rv < 0 ) {
 			unless ( $self->_set_rw_error( $ssl,$rv )) {
 				$self->error("SSL connect attempt failed with unknown error");
 				delete ${*$self}{'_SSL_opening'};
-				${*$self}{'_SSL_opened'} = 1;
-				#DEBUG( "fatal SSL error: $SSL_ERROR" );
+				${*$self}{'_SSL_opened'} = -1;
+				DEBUG(1, "fatal SSL error: $SSL_ERROR" );
 				return $self->fatal_ssl_error();
 			}
 
-			#DEBUG( 'ssl handshake in progress' );
+			DEBUG(2,'ssl handshake in progress' );
 			# connect failed because handshake needs to be completed
 			# if socket was non-blocking or no timeout was given return with this error
 			return if ! defined($timeout);
@@ -340,27 +368,27 @@ sub connect_SSL {
 			if ( $timeout>0 ) {
 				my $vec = '';
 				vec($vec,$self->fileno,1) = 1;
-				#DEBUG( "waiting for fd to become ready: $SSL_ERROR" );
+				DEBUG(2, "waiting for fd to become ready: $SSL_ERROR" );
 				$rv = 
 					$SSL_ERROR == SSL_WANT_READ ? select( $vec,undef,undef,$timeout) :
 					$SSL_ERROR == SSL_WANT_WRITE ? select( undef,$vec,undef,$timeout) :
 					undef;
 			} else {
-				#DEBUG( "handshake failed because no more time" );
+				DEBUG(2,"handshake failed because no more time" );
 				$! = ETIMEDOUT
 			}
 			if ( ! $rv ) {
-				#DEBUG( "handshake failed because socket did not became ready" );
+				DEBUG(2,"handshake failed because socket did not became ready" );
 				# failed because of timeout, return
 				$! ||= ETIMEDOUT;
 				delete ${*$self}{'_SSL_opening'};
-				${*$self}{'_SSL_opened'} = 1;
+				${*$self}{'_SSL_opened'} = -1;
 				$self->blocking(1); # was blocking before
 				return 
 			}
 
 			# socket is ready, try non-blocking connect again after recomputing timeout
-			#DEBUG( "socket ready, retrying connect" );
+			DEBUG(2,"socket ready, retrying connect" );
 			my $now = time();
 			$timeout -= $now - $start;
 			$start = $now;
@@ -368,14 +396,14 @@ sub connect_SSL {
 
 		} elsif ( $rv == 0 ) {
 			delete ${*$self}{'_SSL_opening'};
-			#DEBUG( "connection failed - connect returned 0" );
+			DEBUG(2,"connection failed - connect returned 0" );
 			$self->error("SSL connect attempt failed because of handshake problems" );
-			${*$self}{'_SSL_opened'} = 1;
+			${*$self}{'_SSL_opened'} = -1;
 			return $self->fatal_ssl_error();
 		}
 	}
 
-	#DEBUG( 'ssl handshake done' );
+	DEBUG(2,'ssl handshake done' );
 	# ssl connect successful
 	delete ${*$self}{'_SSL_opening'};
 	${*$self}{'_SSL_opened'}=1;
@@ -417,13 +445,13 @@ sub accept {
 	my $socket = ${*$self}{'_SSL_opening'};
 	if ( ! $socket ) {
 		# underlying socket not done
-		#DEBUG( 'no socket yet' );
+		DEBUG(2,'no socket yet' );
 		$socket = $self->SUPER::accept($class) || return;
-		#DEBUG( 'accept created normal socket '.$socket );
+		DEBUG(2,'accept created normal socket '.$socket );
 	}
 
 	$self->accept_SSL($socket) || return;
-	#DEBUG( 'accept_SSL ok' );
+	DEBUG(2,'accept_SSL ok' );
 
 	return wantarray ? ($socket, getpeername($socket) ) : $socket;
 }
@@ -435,7 +463,7 @@ sub accept_SSL {
 
 	my $ssl;
 	if ( ! ${*$self}{'_SSL_opening'} ) {
-		#DEBUG( 'starting sslifying' );
+		DEBUG(2,'starting sslifying' );
 		${*$self}{'_SSL_opening'} = $socket;
 		my $arg_hash = ${*$self}{'_SSL_arguments'};
 		${*$socket}{'_SSL_arguments'} = { %$arg_hash, SSL_server => 0 };
@@ -459,7 +487,7 @@ sub accept_SSL {
 	$ssl ||= ${*$socket}{'_SSL_object'};
 
 	$SSL_ERROR = undef;
-	#DEBUG( 'calling ssleay::accept' );
+	#DEBUG(2,'calling ssleay::accept' );
 
 	my $timeout = exists $args->{Timeout} 
 		? $args->{Timeout} 
@@ -475,12 +503,12 @@ sub accept_SSL {
 	my $start = defined($timeout) && time();
 	for my $dummy (1) {
 		my $rv = Net::SSLeay::accept($ssl);
-		#DEBUG( 'called ssleay::accept rv='.$rv );
+		DEBUG(3, "Net::SSLeay::accept -> $rv" );
 		if ( $rv < 0 ) {
 			unless ( $socket->_set_rw_error( $ssl,$rv )) {
 				$socket->error("SSL accept attempt failed with unknown error");
 				delete ${*$self}{'_SSL_opening'};
-				${*$socket}{'_SSL_opened'} = 1;
+				${*$socket}{'_SSL_opened'} = -1;
 				return $socket->fatal_ssl_error();
 			}
 
@@ -504,7 +532,7 @@ sub accept_SSL {
 				# failed because of timeout, return
 				$! ||= ETIMEDOUT;
 				delete ${*$self}{'_SSL_opening'};
-				${*$socket}{'_SSL_opened'} = 1;
+				${*$socket}{'_SSL_opened'} = -1;
 				$socket->blocking(1); # was blocking before
 				return 
 			}
@@ -518,12 +546,12 @@ sub accept_SSL {
 		} elsif ( $rv == 0 ) {
 			$socket->error("SSL connect accept failed because of handshake problems" );
 			delete ${*$self}{'_SSL_opening'};
-			${*$socket}{'_SSL_opened'} = 1;
+			${*$socket}{'_SSL_opened'} = -1;
 			return $socket->fatal_ssl_error();
 		}
 	}
 
-	#DEBUG( 'handshake done, socket ready' );
+	DEBUG(2,'handshake done, socket ready' );
 	# socket opened
 	delete ${*$self}{'_SSL_opening'};
 	${*$socket}{'_SSL_opened'} = 1;
@@ -739,7 +767,8 @@ sub close {
 sub stop_SSL {
 	my $self = shift || return _invalid_object();
 	my $stop_args = (ref($_[0]) eq 'HASH') ? $_[0] : {@_};
-	return $self->error("SSL object already closed") unless (${*$self}{'_SSL_opened'});
+	return $self->error("SSL object already closed") 
+		unless (${*$self}{'_SSL_opened'} == 1);
 
 	if (my $ssl = ${*$self}{'_SSL_object'}) {
 		my $shutdown_done;
@@ -813,7 +842,7 @@ sub stop_SSL {
 sub kill_socket {
 	my $self = shift;
 	shutdown($self, 2);
-	$self->close(SSL_no_shutdown => 1) if (${*$self}{'_SSL_opened'});
+	$self->close(SSL_no_shutdown => 1) if (${*$self}{'_SSL_opened'} == 1);
 	delete(${*$self}{'_SSL_ctx'});
 	return;
 }
@@ -864,7 +893,7 @@ sub start_SSL {
 	my $start_handshake = $arg_hash->{SSL_startHandshake};
 	if ( ! defined($start_handshake) || $start_handshake ) {
 		# if we have no callback force blocking mode
-		#DEBUG( "start handshake" );
+		DEBUG(2, "start handshake" );
 		my $blocking = $socket->blocking(1);
 		my $result = ${*$socket}{'_SSL_arguments'}{SSL_server}
 			? $socket->accept_SSL(%to)
@@ -872,7 +901,7 @@ sub start_SSL {
 		$socket->blocking(0) if !$blocking;
 		return $result ? $socket : (bless($socket, $original_class) && ());
 	} else {
-		#DEBUG( "dont start handshake: $socket" );
+		DEBUG(2, "dont start handshake: $socket" );
 		return $socket; # just return upgraded socket 
 	}
 
@@ -885,11 +914,11 @@ sub new_from_fd {
 		(my $mode = $_[0]) =~ tr/+<>//d;
 		shift unless length($mode);
 	}
-	my $handle = IO::Socket::INET->new_from_fd($fd, '+<')
+	my $handle = $ISA[0]->new_from_fd($fd, '+<')
 		|| return($class->error("Could not create socket from file descriptor."));
 
 	# Annoying workaround for Perl 5.6.1 and below:
-	$handle = IO::Socket::INET->new_from_fd($handle, '+<');
+	$handle = $ISA[0]->new_from_fd($handle, '+<');
 
 	return $class->start_SSL($handle, @_);
 }
@@ -1003,14 +1032,14 @@ sub dump_peer_certificate {
 		my $cert = shift;
 		my $scheme = shift || 'none';
 		if ( ! ref($scheme) ) {
-			DEBUG( "scheme=$scheme cert=$cert" );
+			DEBUG(3, "scheme=$scheme cert=$cert" );
 			$scheme = $scheme{$scheme} or croak "scheme $scheme not defined";
 		}
 
 		# get data from certificate
 		my $commonName = $dispatcher{cn}->($cert);
 		my @altNames = $dispatcher{subjectAltNames}->($cert);
-		DEBUG( "identity=$identity cn=$commonName alt=@altNames" );
+		DEBUG(3,"identity=$identity cn=$commonName alt=@altNames" );
 
 		if ( my $sub = $scheme->{callback} ) {
 			# use custom callback
@@ -1127,7 +1156,7 @@ sub get_ssleay_error {
 sub error {
 	my ($self, $error, $destroy_socket) = @_;
 	$error .= Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error());
-	carp $error."\n".$self->get_ssleay_error() if $DEBUG;
+	DEBUG(2, $error."\n".$self->get_ssleay_error());
 	$SSL_ERROR = dualvar( -1, $error );
 	${*$self}{'_SSL_last_err'} = $SSL_ERROR if (ref($self));
 	return;
@@ -1136,7 +1165,8 @@ sub error {
 
 sub DESTROY {
 	my $self = shift || return;
-	$self->close(_SSL_in_DESTROY => 1, SSL_no_shutdown => 1) if (${*$self}{'_SSL_opened'});
+	$self->close(_SSL_in_DESTROY => 1, SSL_no_shutdown => 1) 
+		if (${*$self}{'_SSL_opened'} == 1);
 	delete(${*$self}{'_SSL_ctx'});
 }
 
@@ -1260,7 +1290,7 @@ use constant SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER => 2;
 # it can be blessed.
 sub new {
 	my $class = shift;
-	DEBUG( "$class @_" );
+	#DEBUG( "$class @_" );
 	my $arg_hash = (ref($_[0]) eq 'HASH') ? $_[0] : {@_};
 
 	my $ctx_object = $arg_hash->{'SSL_reuse_ctx'};
@@ -1302,7 +1332,7 @@ sub new {
 
 	if ($arg_hash->{'SSL_check_crl'}) {
 		if (Net::SSLeay::OPENSSL_VERSION_NUMBER() >= 0x0090702f) {
-			Net::SSLeay::X509_STORE_CTX_set_flags(
+			Net::SSLeay::X509_STORE_set_flags(
 				Net::SSLeay::CTX_get_cert_store($ctx),
 				Net::SSLeay::X509_V_FLAG_CRL_CHECK()
 			);
@@ -1371,14 +1401,14 @@ sub new {
 				Net::SSLeay::X509_NAME_oneline(Net::SSLeay::X509_get_subject_name($cert));
 			$error &&= Net::SSLeay::ERR_error_string($error);
 		}
-		DEBUG( "ok=$ok cert=$cert" );
+		DEBUG(3, "ok=$ok cert=$cert" );
 		return $verify_cb->($ok, $ctx_store, $cert, $error);
 	};
 
 	Net::SSLeay::CTX_set_verify($ctx, $verify_mode, $verify_callback);
 
 	$ctx_object = { context => $ctx };
-	DEBUG( "new ctx $ctx" );
+	DEBUG(3, "new ctx $ctx" );
 	$CTX_CREATED_IN_THIS_THREAD{$ctx} = 1;
 
 	if ( my $cache = $arg_hash->{SSL_session_cache} ) {
@@ -1409,14 +1439,14 @@ sub has_session_cache {
 }
 
 
-sub CLONE { %CTX_CREATED_IN_THIS_THREAD = (); DEBUG( "clone!" ) }
+sub CLONE { %CTX_CREATED_IN_THIS_THREAD = (); }
 sub DESTROY {
 	my $self = shift;
 	if ( my $ctx = $self->{context} ) {
-		DEBUG( "free ctx $ctx open=".join( " ",keys %CTX_CREATED_IN_THIS_THREAD ));
+		DEBUG( 3,"free ctx $ctx open=".join( " ",keys %CTX_CREATED_IN_THIS_THREAD ));
 		if ( %CTX_CREATED_IN_THIS_THREAD and 
 			delete $CTX_CREATED_IN_THIS_THREAD{$ctx} ) {
-			DEBUG( "OK free ctx $ctx" );
+			DEBUG( 3,"OK free ctx $ctx" );
 			Net::SSLeay::CTX_free($ctx);
 		}
 	}
@@ -1876,6 +1906,12 @@ is expecting to read from or write to the socket and wants to be satisfied befor
 get to do anything. But with version 0.98 you are better comparing the global exported 
 variable $SSL_ERROR against the exported symbols SSL_WANT_READ and SSL_WANT_WRITE.
 
+=item B<opened()>
+
+This returns false if the socket could not be opened, 1 if the socket could be opened
+and the SSL handshake was successful done and -1 if the underlying IO::Handle is open,
+but the SSL handshake failed.
+
 =item B<< IO::Socket::SSL->start_SSL($socket, ... ) >>
 
 This will convert a glob reference or a socket that you provide to an IO::Socket::SSL
@@ -1978,6 +2014,19 @@ and read/sysread families instead.
 
 =back
 
+=head1 IPv6
+
+Support for IPv6 with IO::Socket::SSL is expected to work and basic testing is done.
+If IO::Socket::INET6 is available it will automatically use it instead of 
+IO::Socket::INET4. 
+
+Please be aware of the associated problems: If you give a name as a host and the
+host resolves to both IPv6 and IPv4 it will try IPv6 first and if there is no IPv6
+connectivity it will fail. 
+
+To avoid these problems you can either force IPv4 by specifying and AF_INET as the
+Domain (this is per socket) or load IO::Socket::SSL with the option 'inet4'
+(This is a global setting, e.g. affects all IO::Socket::SSL objects in the program).
 
 =head1 RETURN VALUES
 
@@ -1990,55 +2039,35 @@ The moral of the story: I<always> check the return values of these functions bef
 using them in any way that you consider meaningful.
 
 
-=head1 IPv6
-
-Support for IPv6 with IO::Socket::SSL is expected to work, but is experimental, as
-none of the author's machines use IPv6 and hence he cannot test IO::Socket::SSL with
-them.  However, a few brave people have used it without incident, so if you wish to
-make IO::Socket::SSL IPv6 aware, pass the 'inet6' option to IO::Socket::SSL when
-calling it (i.e. C<use IO::Socket::SSL qw(inet6);>).  You will need IO::Socket::INET6
-and Socket6 to use this option, and you will also need to write C<use Socket6;> before
-using IO::Socket::SSL.	If you absolutely do not want to use this (or want a quick
-change back to IPv4), pass the 'inet4' option instead.
-
-Currently, there is no support for using IPv4 and IPv6 simultaneously in a single program, 
-but it is planned for a future release.
-
-
 =head1 DEBUGGING
 
 If you are having problems using IO::Socket::SSL despite the fact that can recite backwards
 the section of this documentation labelled 'Using SSL', you should try enabling debugging.	To
-specify the debug level, pass 'debug#' (where # is a number from 0 to 4) to IO::Socket::SSL
-when calling it:
+specify the debug level, pass 'debug#' (where # is a number from 0 to 3) to IO::Socket::SSL
+when calling it. 
+The debug level will also be propagated to Net::SSLeay::trace, see also L<Net::SSLeay>:
 
 =over 4
 
 =item use IO::Socket::SSL qw(debug0);
 
-#No debugging (default).
+No debugging (default).
 
 =item use IO::Socket::SSL qw(debug1);
 
-#Only print out errors.
+Print out errors from IO::Socket::SSL and ciphers from Net::SSLeay.
 
 =item use IO::Socket::SSL qw(debug2);
 
-#Print out errors and cipher negotiation.
+Print also information about call flow from IO::Socket::SSL and progress
+information from Net::SSLeay.
 
 =item use IO::Socket::SSL qw(debug3);
 
-#Print out progress, ciphers, and errors.
-
-=item use IO::Socket::SSL qw(debug4);
-
-#Print out everything, including data.
+Print also some data dumps from IO::Socket::SSL and from Net::SSLeay.
 
 =back
 
-You can also set $IO::Socket::SSL::DEBUG to 0-4, but that's a bit of a mouthful,
-isn't it?
-
 =head1 EXAMPLES
 
 See the 'example' directory.
